#include "crc.h"

#include <fstream>
#include <stdlib.h>

string token_get(const string& s, int& ptr, const char* sep) {
	int start = ptr;
	while (ptr < s.length() && strchr(sep,s[ptr])==0)
		++ptr;
	return string(s,start,ptr-start);
}

void token_skip(const string& s, int& ptr, const char* sep) {
	while (ptr < s.length() && strchr(sep,s[ptr])!=0)
		++ptr;
}

string strip_space(const string& s) {
	string r = s;
	while (r.length() && isspace(r[0]))
		r.erase(0,1);
	while (r.length() && isspace(r[r.length()-1]))
		r.erase(r.length()-1,1);
	return r;
}

static void crc_scan(const crc_list* cl, istream& f) {
	while (!f.eof()) {
		string s;
		getline(f,s,'\n');

		if (s.length() && s[0]!='#') {
			int ptr = 0;

			crc_info ci;

			token_skip(s,ptr," ");
			string first = token_get(s,ptr," =");
			if (first.length() && first[0]!='[') {
				ci.crc = strtoul(first.c_str(),0,16);
				if (ci.crc) {
					token_skip(s,ptr," =");
					ci.description = strip_space(token_get(s,ptr,"|"));
					token_skip(s,ptr," ");
					if (ptr < s.length() && s[ptr]=='|') ++ptr;
					token_skip(s,ptr," ");
					ci.year = strip_space(token_get(s,ptr,"|"));
					token_skip(s,ptr," ");
					if (ptr < s.length() && s[ptr]=='|') ++ptr;
					token_skip(s,ptr," ");
					ci.manufacturer = strip_space(token_get(s,ptr,"|"));
					cl->bag.insert(ci);
				}
			}
		}
	}
}

void crc_scan(crc_list_set& cls, const game_set& gs, const string& crc_dir) {
	for(game_set::const_iterator i=gs.begin();i!=gs.end();++i) {
		string path = crc_dir + i->name_get() + ".crc";
		ifstream f(path.c_str());
		if (f.good()) {
			crc_list cl;
			cl.name = i->name_get();
			pair<crc_list_set::const_iterator,bool> j = cls.insert(cl);
			crc_scan(&*j.first,f);
			f.close();
		}
	}
}
