/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __MENU_H
#define __MENU_H

#include "game.h"

enum emulator_t {
	emulator_mame,
	emulator_mess
};

// Type of sorting
enum game_sort_t {
	sort_by_group,
	sort_by_name,
	sort_by_root_name,
	sort_by_time,
	sort_by_year,
	sort_by_manufacturer,
	sort_by_type,
	sort_by_size
};

inline bool sort_by_root_name_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_leveldesc_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_name_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_manufacturer_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_manufacturer_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_year_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_year_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_time_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_time_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_group_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_group_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_type_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_type_less,pgame_by_desc_less,pgame_by_name_less);
}

inline bool sort_by_size_func(const game* A, const game* B) {
	return pgame_combine_less(A,B,pgame_by_size_less,pgame_by_desc_less,pgame_by_name_less);
}

typedef	bool (*pgame_sort_func)(const game*, const game*);

typedef set<const game*,pgame_sort_func> pgame_sort_set;

enum tristate_t {
	include,
	exclude,
	exclude_not
};

// List mode
enum show_t {
	mode_list,
	mode_tile_3x2,
	mode_tile_4x3,
	mode_tile_5x4
};

// Type of image to display
enum preview_t {
	preview_snap,
	preview_flyers
};

struct menu_state {
	emulator_t emulator;

	group_container include_group; // gruppi di inclusione
	type_container include_type; // gruppi tipo

	tristate_t exclude_neogeo;
	tristate_t exclude_clone;
	tristate_t exclude_bad;
	tristate_t exclude_missing;
	tristate_t exclude_vector;

	bool crc_scan; // use the crc info to get game informations (MESS)

	preview_t preview; // type of preview selected
	string software_path; // software directories
	string snap_create_path; // snapshot create directory
	string snap_path; // snapshot directories
	string flyers_path; // flyers directories
	string crc_path; // crc directory

	int pos_base; // posizione di inizio schermo
	int pos_rel; // posizione relativa del cursore
	const game* pgame; // gioco alla posizione del cursore, ==0 se nessuno
	const game* pclone; // clone scelto, ==0 se nessuno
	string fast; // stringa per selezione veloce

	game_sort_t sort_mode; // sort mode

	show_t mode; // list mode

	string last_game; // last game selected

	unsigned idle_start;
	unsigned idle_start_rep;
	unsigned idle_move;
	unsigned idle_move_rep;

	merge_t merge; // merge type

	unsigned exit_count; // number of times the exit buttun need to be pressed
};

void run_help();
int run_menu(const game_set& gar, menu_state& rs, bool flipxy);
void run_sort(menu_state& rs);
void run_exclude(menu_state& rs);
void run_type(menu_state& rs, const type_container& typ);
void run_type_move(menu_state& rs, const type_container& typ);
void run_group(menu_state& rs, const group_container& grp);
void run_group_move(menu_state& rs, const group_container& grp);
void run_clone(menu_state& rs);

bool color_in(const string& s);
void color_out_all(ostream& os);

#endif
