/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __TEXT_H
#define __TEXT_H

#include <stdlib.h>
#include <time.h>

#include <string>

#define TEXT_KEY_NONE 0
#define TEXT_KEY_UP 1
#define TEXT_KEY_DOWN 2
#define TEXT_KEY_LEFT 3
#define TEXT_KEY_RIGHT 4
#define TEXT_KEY_ENTER 5
#define TEXT_KEY_ESC 6
#define TEXT_KEY_SPACE 7
#define TEXT_KEY_MODE 8
#define TEXT_KEY_HOME 9
#define TEXT_KEY_END 10
#define TEXT_KEY_PGUP 11
#define TEXT_KEY_PGDN 12
#define TEXT_KEY_HELP 13
#define TEXT_KEY_GROUP 14
#define TEXT_KEY_TYPE 15
#define TEXT_KEY_EXCLUDE 16
#define TEXT_KEY_SORT 17
#define TEXT_KEY_SETGROUP 21
#define TEXT_KEY_SETTYPE 22
#define TEXT_KEY_RUN_CLONE 23
#define TEXT_KEY_IDLE_0 24
#define TEXT_KEY_IDLE_1 25
#ifdef __SNAPSHOT__
#define TEXT_KEY_SNAPSHOT 26
#endif
#define TEXT_KEY_INS 27
#define TEXT_KEY_DEL 28

bool text_init(const char* cfg);
void text_done();
bool text_sub_init(const char* font, unsigned orientation, double gamma, double brightness, unsigned idle_0, unsigned idle_0_rep,unsigned idle_1, unsigned idle_1_rep);
void text_sub_done();

void text_clear();
void text_clear(int x, int y, int dx, int dy, int color);
void text_box(int x, int y, int dx, int dy, int color);

void text_put(int x, int y, char c, int color);
unsigned text_put_width(char c);
void text_put(int x, int y, const string& s, int color);
unsigned text_put_width(const string& s);

void text_backdrop_init(unsigned back_color, unsigned Amac);
void text_backdrop_clear(int back_index, bool highlight);
void text_backdrop_clear();
void text_backdrop_set(int back_index, const char* file, bool highlight);
void text_backdrop_set(int back_index, int x, int y, int dx, int dy);
void text_backdrop_update();

unsigned text_getkey();
int text_keypressed();
void text_idle_repeat_reset();
void text_idle_time_reset();

int text_font_dx_get();
int text_font_dy_get();

int text_dx_get();
int text_dy_get();

void text_key_out_all(ostream& os);
bool text_key_in(const string& s);

#endif
