/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "vbeline.h"
#include "vbe.h"

#include "../card/card.h"
#include "../card/vbe3.h"
#include "../card/laguna.h"
#include "../card/s6326.h"
#include "../card/savage.h"
#include "../card/tdfx.h"
#include "../card/ati.h"
#include "../card/s3.h"
#include "../card/cirrus.h"
#include "../card/matrox.h"
#include "../card/r128.h"
#include "../card/neomagic.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/***************************************************************************/
/* Interface */

vbeline_driver cards[]=
{
	{"laguna", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_LAGUNA, detectLaguna, setLaguna, resetLaguna }, /* Laguna driver */
	{"3dfx", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_3DFX, tdfx_detect, tdfx_set, tdfx_reset }, /* 3dfx driver */
	{"savage", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_SAVAGE, detectSavage, setSavage, resetSavage }, /* Savage driver */
	{"sis6326", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_SIS6326, detectSiS6326, setSiS6326, resetSiS6326 }, /* SiS6326 driver */
	{"matrox", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_MATROX, matrox_detect, matrox_set, matrox_reset }, /* Matrox driver */
	{"r128", 0, 1, VIDEO_OPTION_VBELINE_DRIVER_R128, r128_detect, r128_set, r128_reset }, /* R128 driver */
	{"neomagic", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_NEOMAGIC, neomagic_detect, neomagic_set, neomagic_reset }, /* NeoMagic driver */
	{"s3", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_S3, s3_detect, s3_set, s3_reset }, /* S3 driver */
	{"ati", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_ATI, detectATI, setATI, resetATI } , /* ATI driver */
	{"cirrus", 0, 0, VIDEO_OPTION_VBELINE_DRIVER_CIRRUS, cirrus_detect, cirrus_set, cirrus_reset }, /* Cirrus driver */
	{"vbe3", 1, 0, VIDEO_OPTION_VBELINE_DRIVER_VBE3, vbe3_detect, vbe3_set, vbe3_reset }, /* VBE3 driver */
	{0,0,0,0,0,0,0}
};

#define is_vga_mode(mode) ((mode) < 0x100)

int card_mode_callback(CRTCPARAM *cp) {
	unsigned de,bs,rs,re,be,tt;

	cp->bits = vbeline_state.vbe_info_request.BitsPerPixel;
	cp->original_width = vbeline_state.vbe_info_request.XResolution;
	cp->original_height = vbeline_state.vbe_info_request.YResolution;

	video_crtc_realize_h(&vbeline_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->HDisp = de;
	cp->HBStart = bs;
	cp->HSStart = rs;
	cp->HSEnd = re;
	cp->HBEnd = be;
	cp->HTotal = tt;

	video_crtc_realize_v(&vbeline_state.crtc_request,&de,&bs,&rs,&re,&be,&tt,1);
	cp->VDisp = de;
	cp->VBStart = bs;
	cp->VSStart = rs;
	cp->VSEnd = re;
	cp->VBEnd = be;
	cp->VTotal = tt;

	/* VSyncMAME driver requires effective vertical values */
	if (vbeline_state.crtc_request.doublescan) {
		cp->VDisp *= 2;
		cp->VBStart *= 2;
		cp->VSStart *= 2;
		cp->VSEnd *= 2;
		cp->VBEnd *= 2;
		cp->VTotal *= 2;
	}
	if (vbeline_state.crtc_request.interlaced) {
		cp->VDisp /= 2;
		cp->VBStart /= 2;
		cp->VSStart /= 2;
		cp->VSEnd /= 2;
		cp->VBEnd /= 2;
		cp->VTotal /= 2;
	}

	cp->width = vbeline_state.crtc_request.hde;
	cp->height = vbeline_state.crtc_request.vde;
	cp->hpolarity = vbeline_state.crtc_request.hsync;
	cp->vpolarity = vbeline_state.crtc_request.vsync;
	cp->doublescan = vbeline_state.crtc_request.doublescan;
	cp->interlace = vbeline_state.crtc_request.interlaced;
	cp->interlaceratio = 50;

	cp->dotclockHz = vbeline_state.crtc_request.pixelclock;

	/* vga mode 13h requires double horizontal value */
	if (vbeline_state.mode_request == 0x13) {
		cp->HDisp *= 2;
		cp->HBStart *= 2;
		cp->HSStart *= 2;
		cp->HSEnd *= 2;
		cp->HBEnd *= 2;
		cp->HTotal *= 2;
		cp->dotclockHz *= 2;
	}

	/* vga mode 3h requires CRTC values computed on font of width 8 */
	if (vbeline_state.mode_request == 0x3 && (cp->HDisp % 9 == 0)) {
		cp->HDisp = cp->HDisp * 8 / 9;
		cp->HBStart = cp->HBStart * 8 / 9;
		cp->HSStart = cp->HSStart * 8 / 9;
		cp->HSEnd = cp->HSEnd * 8 / 9;
		cp->HBEnd = cp->HBEnd * 8 / 9;
		cp->HTotal = cp->HTotal * 8 / 9;
	}

	video_log("card: expected vert clock: %.2f Hz\n", (double)cp->dotclockHz / (cp->VTotal * cp->HTotal) );
	video_log_modeline_b("card: mode_set", cp->dotclockHz, cp->HDisp, cp->HBStart, cp->HSStart, cp->HSEnd, cp->HBEnd, cp->HTotal,cp->VDisp, cp->VBStart, cp->VSStart, cp->VSEnd, cp->VBEnd, cp->VTotal, cp->hpolarity, cp->vpolarity, cp->doublescan, cp->interlace);

	return 1;
}

void card_log(const char* text,...) {
	va_list arg;
	va_start(arg,text);
	video_log_va(text,arg);
	va_end(arg);
}

/***************************************************************************/

unsigned FAST (*vbeline_write_selector)(void);

/* Return the offset for accessing in writing the video memory */
unsigned FAST (*vbeline_write_line)(unsigned y);

void vbeline_init(void) {
	unsigned i;

	assert( !vbeline_state.active );

	/* VBE (any) is required */
	if (!vbe_active()) {
		video_log("vbeline: a vbe bios is required\n");
		return;
	}

	if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_NONE)
		return;

	vbeline_state.driver = 0;
	for(i=0;cards[i].name;++i) {
		if (video_option_vbeline_driver_get() == VIDEO_OPTION_VBELINE_DRIVER_AUTO
			|| video_option_vbeline_driver_get() == cards[i].type) {
			if (!cards[i].require_vbe3 || vbe_state.info.VESAVersion >= 0x300) {
				if (cards[i].detect && cards[i].detect() > 0) {
					video_log("vbeline: found driver %s\n",cards[i].name);
					vbeline_state.driver = cards + i;
					break;
				}
			}
		}
	}

	vbeline_state.active = vbeline_state.driver != 0;
}

void vbeline_done(void) {
	assert( vbeline_active() );
	assert( !vbeline_mode_active() );

	if (vbeline_state.driver->reset)
		vbeline_state.driver->reset();

	vbeline_state.active = 0;
}

static int vbe_mode_preset(vbe_ModeInfoBlock* info, vbe_CRTCInfoBlock* crtc) {
	CRTCPARAM cp;

	if (!card_mode_callback(&cp))
		return -1;

	if ((info->ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that the selected mode isn't avaliable");
		return -1;
	}

	if (cp.interlace && (info->ModeAttributes & vbeMdInterlace)==0) {
		video_error_description_set("VBE report that the selected mode doesn't support interlace");
		return -1;
	}

	if (cp.doublescan && (info->ModeAttributes & vbeMdDoubleScan)==0) {
		video_error_description_set("VBE report that the selected mode doesn't support doublescan");
		return -1;
	}

	if (cp.interlace && cp.doublescan) {
		video_error_description_set("VBE doesn't support interlaced and doublescan at the same time");
		return -1;
	}

	memset(crtc,0,sizeof(vbe_CRTCInfoBlock));

	/* if the resulting mode is smaller and if later sync values are set */
	/* use fake value to prevent a possible mode set failure if the */
	/* specified values are smaller than the fixed resolution */

	if (cp.HDisp < info->XResolution) {
		/* use fake values, correct values are set later */
		crtc->HorizontalSyncStart = info->XResolution * 41 / 40;
		crtc->HorizontalSyncEnd = info->XResolution * 10 / 9;
		crtc->HorizontalTotal = info->XResolution * 5 / 4;
	} else {
		crtc->HorizontalTotal = cp.HTotal;
		crtc->HorizontalSyncStart = cp.HSStart;
		crtc->HorizontalSyncEnd = cp.HSEnd;
	}
	if (cp.VDisp < info->YResolution) {
		/* use fake values, correct values are set later */
		crtc->VerticalSyncStart = info->YResolution * 31 / 30;
		crtc->VerticalSyncEnd = info->YResolution * 21 / 20;
		crtc->VerticalTotal = info->YResolution * 16 / 15;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (cp.doublescan) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	} else {
		crtc->VerticalTotal = cp.VTotal;
		crtc->VerticalSyncStart = cp.VSStart;
		crtc->VerticalSyncEnd = cp.VSEnd;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (cp.interlace) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	}

	crtc->Flags = 0;
	if (cp.doublescan)
		crtc->Flags |= vbeCRTCDoubleScan;
	if (cp.interlace)
		crtc->Flags |= vbeCRTCInterlaced;
	if (cp.hpolarity)
		crtc->Flags |= vbeCRTCHorizontalSyncNegative;
	if (cp.vpolarity)
		crtc->Flags |= vbeCRTCVerticalSyncNegative;

	crtc->PixelClock = cp.dotclockHz;
	crtc->RefreshRate = (long)crtc->PixelClock * 100 / ((long)crtc->HorizontalTotal * crtc->VerticalTotal);

	video_log("vbeline: mode_preset %dx%d (%.2f %d %d %d %d %d %d)\n",(unsigned)info->XResolution,(unsigned)info->YResolution,(double)crtc->PixelClock/1E6,(unsigned)crtc->HorizontalSyncStart,(unsigned)crtc->HorizontalSyncEnd,(unsigned)crtc->HorizontalTotal,(unsigned)crtc->VerticalSyncStart,(unsigned)crtc->VerticalSyncEnd,(unsigned)crtc->VerticalTotal);

	return 0;
}

int vbeline_mode_set(vbeline_video_mode* mode) {
	assert( vbeline_active() );
	assert( !vbeline_mode_active() );

	video_log_modeline_m("vbeline: mode_set", mode->mode, mode->crtc.pixelclock, mode->crtc.hde, mode->crtc.hrs, mode->crtc.hre, mode->crtc.ht, mode->crtc.vde, mode->crtc.vrs, mode->crtc.vre, mode->crtc.vt, mode->crtc.hsync, mode->crtc.vsync, mode->crtc.doublescan, mode->crtc.interlaced );

	vbeline_state.mode_request = mode->mode;
	vbeline_state.crtc_request = mode->crtc;

	if (vbeline_state.driver->require_mode_save) {
		video_log("vbeline: save video board state\n");
		if (vbe_save_state(&vbeline_state.state_size,&vbeline_state.state_ptr) != 0) {
			return -1;
		}
	}

	if (is_vga_mode(vbeline_state.mode_request)) {
		if (vga_as_vbe_mode_info_get(&vbeline_state.vbe_info_request, vbeline_state.mode_request)!=0)
			return -1;
	} else {
		if (vbe_mode_info_get(&vbeline_state.vbe_info_request, vbeline_state.mode_request)!=0)
			return -1;
	}

	if (video_option_vbeline_clock_multiplier_get()!=1.0) {
		video_log("vbeline: clock multiplier: %.2f\n",video_option_vbeline_clock_multiplier_get());
		vbeline_state.crtc_request.pixelclock *= video_option_vbeline_clock_multiplier_get();
	}

	if (vbeline_state.driver->require_vbe3) {
		if (is_vga_mode(vbeline_state.mode_request)) {
			video_error_description_set("VBE3 ins't supported for VGA modes");
			return -1;
		}
		if (vbe_pixelclock_get(&vbeline_state.crtc_request.pixelclock,vbeline_state.mode_request)!=0) {
			return -1;
		}
		if (vbe_mode_preset(&vbeline_state.vbe_info_request,&vbeline_state.vbe3_crtc_request)!=0) {
			return -1;
		}
		if (vbe_mode_set(vbeline_state.mode_request,&vbeline_state.vbe3_crtc_request)!=0) {
			return -1;
		}
	} else {
		if (is_vga_mode(vbeline_state.mode_request)) {
			if (vga_mode_set(vbeline_state.mode_request,0,0)!=0) {
				return -1;
			}
		} else {
			if (vbe_mode_set(vbeline_state.mode_request,0)!=0) {
				return -1;
			}
		}
	}

	if (is_vga_mode(vbeline_state.mode_request)) {
		if (vbeline_state.mode_request == 0x13) {
			/* convert to unchained */
			if (vbeline_state.crtc_request.hde * vbeline_state.crtc_request.vde > 0x10000)
				vga_chained_set(0);
			/* set the scanline length */
			vga_scanline_set(vbeline_state.crtc_request.hde);
		}
	} else {
		if (vbe_scanline_pixel_request(vbeline_state.crtc_request.hde) != 0) {
			vbe_mode_done();
			return -1;
		}
	}

	if (vbeline_state.driver->set() <= 0) {
		video_log("vbeline: error in the video driver mode set\n");
		vbe_mode_done();
		video_error_description_set("Error in the video driver mode set");
		return -1;
	}

	if (is_vga_mode(vbeline_state.mode_request)) {
		/* adjust resolution */
		vbeline_state.size_x = vga_state.mode_info.size_x = vbeline_state.crtc_request.hde;
		vbeline_state.size_y = vga_state.mode_info.size_y = vbeline_state.crtc_request.vde;
		/* adjust virtual size */
		vbeline_state.virtual_x = vga_state.mode_info.virtual_x;
		vbeline_state.virtual_y = vga_state.mode_info.virtual_y;
		/* write handler */
		vbeline_write_selector = vga_write_selector;
		vbeline_write_line = vga_write_line;
	} else {
		/* adjust resolution */
		vbeline_state.size_x = vbe_state.size_x = vbeline_state.crtc_request.hde;
		vbeline_state.size_y = vbe_state.size_y = vbeline_state.crtc_request.vde;
		/* adjust virtual size */
		vbeline_state.virtual_x = vbe_state.virtual_x;
		vbeline_state.virtual_y = vbe_state.virtual_y;
		/* write handler */
		vbeline_write_selector = vbe_write_selector;
		vbeline_write_line = vbe_write_line;
	}

	vbeline_state.mode_active = 1;

	return 0;
}

void vbeline_mode_done(void) {
	assert( vbeline_active() );
	assert( vbeline_mode_active() );

	if (vbeline_state.driver->require_mode_save) {
		video_log("vbeline: restore video board state\n");
		vbe_restore_state(vbeline_state.state_size,vbeline_state.state_ptr);
	}

	if (vbe_mode_active())
		vbe_mode_done();
	else
		vga_mode_done();

	vbeline_state.mode_active = 0;
}

#define CARD_PIXELCLOCK_DELTA (10*1000)
#define CARD_PIXELCLOCK_MAX (250*1000*1000)
#define CARD_PIXELCLOCK_MIN (1*1000*1000)

int vbeline_pixelclock_getnext(unsigned* pixelclock, unsigned mode) {
	assert( vbeline_active() );
	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getnext(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock + CARD_PIXELCLOCK_DELTA < CARD_PIXELCLOCK_MAX)
			*pixelclock += CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

int vbeline_pixelclock_getpred(unsigned* pixelclock, unsigned mode) {
	assert( vbeline_active() );
	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		return vbe_pixelclock_getpred(pixelclock,mode);
	} else {
		/* assume all clocks avaliable */
		if (*pixelclock > CARD_PIXELCLOCK_MIN - CARD_PIXELCLOCK_DELTA)
			*pixelclock -= CARD_PIXELCLOCK_DELTA;
		return 0;
	}
}

/* Import information for one video mode
	mode->vbeline video mode
	Return 0 if successful
*/
int vbeline_mode_import(video_mode* mode) {
	vbe_ModeInfoBlock info;

	if (!vbeline_active()) {
		video_error_description_set("VBE CRTC programming isn't avaliable on VBE version prior of 3 or without the dedicated driver");
		return -1;
	}
	if (vbeline_state.driver->require_vbe3 && is_vga_mode(mode->driver.vbeline.mode)) {
		video_error_description_set("VBE3 doesn't support VGA modes");
		return -1;
	}
	if (is_vga_mode(mode->driver.vbeline.mode)) {
		if (vga_as_vbe_mode_info_get(&info, mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	} else {
		if (vbe_mode_info_get(&info, mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	}
	if ((info.ModeAttributes & vbeMdAvailable) == 0) {
		video_error_description_set("VBE report that mode %d is not avaliable", mode->driver.vbe.mode);
		return -1;
	}
	if ((info.ModeAttributes & vbeMdGraphMode) == 0) {
		if (info.MemoryModel!=vbeMemTXT) {
			video_error_description_set("Unsupported memory model");
			return -1;
		}
		// TODO
		video_error_description_set("Text modes are unsupported");
		return -1;
	} else {
		if ((mode->driver.vbeline.mode & vbeLinearBuffer) != 0) {
			if (vbe_state.info.VESAVersion < 0x200) {
				video_error_description_set("Linear frame buffer isn't available on VBE version prior 2");
				return -1;
			}
			if ((info.ModeAttributes & vbeMdLinear) == 0) {
				video_error_description_set("Linear frame buffer isn't available in mode %x", mode->driver.vbeline.mode);
				return -1;
			}
		} else {
			if ((info.ModeAttributes & vbeMdNonBanked) != 0) {
				video_error_description_set("Banked frame buffer isn't available in mode %x",mode->driver.vbeline.mode);
				return -1;
			}
		}
		/* 8,15,16 bits per pixel */
		if (info.BitsPerPixel!=8 && info.BitsPerPixel!=15 && info.BitsPerPixel!=16) {
			video_error_description_set("Unsupported bits per pixel size");
			return -1;
		}
		/* Packed or RGB memory model */
		if (info.MemoryModel!=vbeMemPK && info.MemoryModel!=vbeMemRGB) {
			video_error_description_set("Unsupported memory model");
			return -1;
		}
		/* Non planar mode */
		if (info.NumberOfPlanes!=1) {
			video_error_description_set("Unsupported number of planes");
			return -1;
		}
	}

	if (vbeline_state.driver->type == VIDEO_OPTION_VBELINE_DRIVER_VBE3) {
		if (vbe_pixelclock_get(&mode->driver.vbeline.crtc.pixelclock,mode->driver.vbeline.mode)!=0) {
			return -1;
		}
	} else {
		/* assume all clocks avaliable */
	}

	mode->flags = VIDEO_FLAGS_DRIVER_VBELINE |
		VIDEO_FLAGS_ASYNC_SETPAGE |
		(mode->flags & VIDEO_FLAGS_USER_MASK);
	if ((info.ModeAttributes & vbeMdTripleBuffer) != 0)
		mode->flags |= VIDEO_FLAGS_SYNC_SETPAGE;
	if (info.MemoryModel==vbeMemTXT)
		mode->flags |= VIDEO_FLAGS_INDEX_TEXT | VIDEO_FLAGS_TYPE_TEXT;
	if (info.MemoryModel==vbeMemPK)
		mode->flags |= VIDEO_FLAGS_INDEX_PACKED | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (info.MemoryModel==vbeMemRGB)
		mode->flags |= VIDEO_FLAGS_INDEX_RGB | VIDEO_FLAGS_TYPE_GRAPHICS;
	if (mode->driver.vbe.mode & vbeLinearBuffer)
		mode->flags |= VIDEO_FLAGS_LINEAR;

	mode->size_x = mode->driver.vbeline.crtc.hde;
	mode->size_y = mode->driver.vbeline.crtc.vde;
	mode->virtual_x = 0;
	mode->virtual_y = 0;
	mode->vert_freq = (double)mode->driver.vbeline.crtc.pixelclock / (mode->driver.vbeline.crtc.ht * mode->driver.vbeline.crtc.vt);
	if (mode->driver.vbeline.crtc.doublescan)
		mode->vert_freq /= 2;
	if (mode->driver.vbeline.crtc.interlaced)
		mode->vert_freq *= 2;
	mode->horz_freq = (double)mode->driver.vbeline.crtc.pixelclock / mode->driver.vbeline.crtc.ht;
	mode->bits = info.BitsPerPixel;
	mode->aspect_x = 4 * mode->size_y;
	mode->aspect_y = 3 * mode->size_x;
	return 0;
}

/* the mode number is set later */
static const char* VBELINE_DEFAULT[] = {

/* step 16 - 240 256 272 288 304 320 336 352 368 384 400 416 432 448 464 480 */

/***************************************************************************/
/* PC MONITOR AT 31.5 kHz */

/* computed */
"pc_31.5_256x192 vbeline(0 10.0701 256 264 304 320 192 196 197 210 -hsync -vsync doublescan)",
"pc_31.5_272x204 vbeline(0 10.5736 272 272 312 336 204 208 209 223 -hsync -vsync doublescan)",
"pc_31.5_288x216 vbeline(0 11.3288 288 296 336 360 216 220 221 236 -hsync -vsync doublescan)",
"pc_31.5_304x228 vbeline(0 11.8323 304 312 352 376 228 232 233 249 -hsync -vsync doublescan)",
"pc_31.5_320x240 vbeline(0 12.5876 320 328 376 400 240 245 246 262 -hsync -vsync doublescan)",
"pc_31.5_336x252 vbeline(0 13.0911 336 344 392 416 252 258 259 276 -hsync -vsync doublescan)",
"pc_31.5_352x264 vbeline(0 13.8464 352 360 416 440 264 270 271 289 -hsync -vsync doublescan)",
"pc_31.5_368x276 vbeline(0 14.3499 368 376 432 456 276 282 283 302 -hsync -vsync doublescan)",
"pc_31.5_384x288 vbeline(0 15.1051 384 392 448 480 288 294 295 315 -hsync -vsync)",
"pc_31.5_400x300 vbeline(0 15.6086 400 408 464 496 300 306 307 328 -hsync -vsync)",
"pc_31.5_416x312 vbeline(0 16.3639 416 424 488 520 312 318 320 341 -hsync -vsync)",
"pc_31.5_432x324 vbeline(0 16.8674 432 440 504 536 324 330 332 354 -hsync -vsync)",
"pc_31.5_448x336 vbeline(0 17.6226 448 456 528 560 336 343 344 367 -hsync -vsync)",
"pc_31.5_464x348 vbeline(0 18.1261 464 472 544 576 348 356 357 381 -hsync -vsync)",
"pc_31.5_480x360 vbeline(0 18.8814 480 488 560 600 360 368 369 394 -hsync -vsync)",
"pc_31.5_496x372 vbeline(0 19.3849 496 504 576 616 372 380 381 407 -hsync -vsync)",
"pc_31.5_512x384 vbeline(0 20.1402 512 528 600 640 384 392 394 420 -hsync -vsync)",
"pc_31.5_576x432 vbeline(0 22.6577 576 592 680 720 432 441 442 472 -hsync -vsync)",
"pc_31.5_640x480 vbeline(0 25.1752 640 656 752 800 480 490 492 525 -hsync -vsync)",
"pc_31.5_672x504 vbeline(0 26.434 672 688 792 840 504 514 516 551 -hsync -vsync)",

/***************************************************************************/
/* PC MONITOR AT 35.16 kHz */

/* "pc_35.16_352x264 vbeline(0 15.81 352 369 401 449 264 274 276 284 doublescan -hsync -vsync)", */
"pc_35.16_368x276 vbeline(0 16.41 368 377 417 465 276 286 288 296 doublescan -hsync -vsync)",
"pc_35.16_384x288 vbeline(0 17.50 384 401 441 497 288 298 300 308 doublescan -hsync -vsync)",
"pc_35.16_400x300 vbeline(0 18.06 400 417 457 513 300 311 313 321 doublescan -hsync -vsync)",
"pc_35.16_432x324 vbeline(0 19.20 432 449 489 545 324 336 338 345 doublescan -hsync -vsync)",
"pc_35.16_800x600 vbeline(0 36.12 800 835 915 1027 600 614 617 642 -hsync -vsync)",

/***************************************************************************/
/* PC MONITOR AT 48.0 kHz */

"pc_48.0_512x384 vbeline(0 31.10 512 520 552 648 384 385 390 400 doublescan -hsync -vsync)",
"pc_48.0_1024x768 vbeline(0 61.44 1024 1040 1128 1280 768 770 780 800 -hsync -vsync)",

/***************************************************************************/
/* PC MONITOR AT 37.0 kHz */

"pc_37.0_1024x768 vbeline(0 47.39 1024 1040 1128 1280 768 770 780 800 -hsync -vsync interlaced)",

/***************************************************************************/
/* PAL TV */

/* SMALL resolutions, computed */
"pal_small_368x276 vbeline(0 7.375 368 384 424 472 276 285 288 312 -hsync -vsync)",
"pal_small_736x276 vbeline(0 14.75 736 776 848 944 276 285 288 312 -hsync -vsync)",
"pal_small_736x552 vbeline(0 14.75 736 776 848 944 552 570 576 625 -hsync -vsync interlaced)",

/* MAX resolutions, computed */
"pal_384x288 vbeline(0 7.375 384 392 432 472 288 291 294 312 -hsync -vsync)",
"pal_768x288 vbeline(0 14.75 768 792 864 944 288 291 294 312 -hsync -vsync)",
"pal_768x576 vbeline(0 14.75 768 792 864 944 576 582 588 625 -hsync -vsync interlaced)",

/***************************************************************************/
/* NTSC TV */

/* SMALL resolutions, computed */
"ntsc_small_304x228 vbeline(0 6.16773 304 328 352 392 228 238 241 262 -hsync -vsync)",
"ntsc_small_608x228 vbeline(0 12.2096 608 640 696 776 228 238 241 262 -hsync -vsync)",
"ntsc_small_608x456 vbeline(0 12.2096 608 640 696 776 456 476 482 525 -hsync -vsync interlaced)",

/* MAX resolutions, computed */
"ntsc_320x242 vbeline(0 6.04186 320 328 352 384 242 245 248 262 -hsync -vsync)",
"ntsc_640x242 vbeline(0 12.2096 640 664 720 776 242 245 248 262 -hsync -vsync)",
"ntsc_640x484 vbeline(0 12.2096 640 664 720 776 484 490 496 525 -hsync -vsync interlaced)",

/***************************************************************************/
/* Arcade 15kHz */

/* computed */
"arcade_15.75_256x192 vbeline(0 5.292 256 272 304 336 192 205 207 232 -hsync -vsync)",
"arcade_15.75_512x192 vbeline(0 10.584 512 552 600 672 192 205 207 232 -hsync -vsync)",
"arcade_15.75_272x204 vbeline(0 5.67 272 296 320 360 204 218 221 247 -hsync -vsync)",
"arcade_15.75_544x204 vbeline(0 11.214 544 584 632 712 204 218 221 247 -hsync -vsync)",
"arcade_15.75_288x216 vbeline(0 5.922 288 312 336 376 216 231 233 261 -hsync -vsync)",
"arcade_15.75_576x216 vbeline(0 11.97 576 624 680 760 216 231 233 261 -hsync -vsync)",
"arcade_15.75_304x228 vbeline(0 6.3 304 328 360 400 228 244 247 276 -hsync -vsync)",
"arcade_15.75_608x228 vbeline(0 12.6 608 656 712 800 228 244 247 276 -hsync -vsync)",
"arcade_15.75_320x240 vbeline(0 6.678 320 344 376 424 240 256 259 290 -hsync -vsync)",
"arcade_15.75_640x240 vbeline(0 13.23 640 688 752 840 240 256 259 290 -hsync -vsync)",
"arcade_15.75_336x252 vbeline(0 6.93 336 360 392 440 252 270 273 305 -hsync -vsync)",
"arcade_15.75_672x252 vbeline(0 13.86 672 720 784 880 252 270 273 305 -hsync -vsync)",
"arcade_15.75_352x264 vbeline(0 7.308 352 384 416 464 264 283 286 320 -hsync -vsync)",
"arcade_15.75_704x264 vbeline(0 14.616 704 760 832 928 264 283 286 320 -hsync -vsync)",
"arcade_15.75_368x276 vbeline(0 7.56 368 392 432 480 276 295 299 334 -hsync -vsync)",
"arcade_15.75_736x276 vbeline(0 15.246 736 792 864 968 276 295 299 334 -hsync -vsync)",
"arcade_15.75_384x288 vbeline(0 7.938 384 416 448 504 288 309 312 349 -hsync -vsync)",
"arcade_15.75_768x288 vbeline(0 15.876 768 824 904 1008 288 309 312 349 -hsync -vsync)",
"arcade_15.75_400x300 vbeline(0 8.316 400 432 472 528 300 321 325 363 -hsync -vsync)",
"arcade_15.75_800x300 vbeline(0 16.506 800 856 936 1048 300 321 325 363 -hsync -vsync)",
"arcade_15.75_416x312 vbeline(0 8.568 416 448 488 544 312 334 338 378 -hsync -vsync interlaced)",
"arcade_15.75_832x312 vbeline(0 17.262 832 896 976 1096 312 334 338 378 -hsync -vsync interlaced)",
"arcade_15.75_432x324 vbeline(0 8.946 432 464 504 568 324 347 350 392 -hsync -vsync interlaced)",
"arcade_15.75_864x324 vbeline(0 17.892 864 928 1016 1136 324 347 350 392 -hsync -vsync interlaced)",
"arcade_15.75_448x336 vbeline(0 9.324 448 488 528 592 336 360 364 407 -hsync -vsync interlaced)",
"arcade_15.75_896x336 vbeline(0 18.522 896 968 1048 1176 336 360 364 407 -hsync -vsync interlaced)",
"arcade_15.75_464x348 vbeline(0 9.576 464 496 544 608 348 372 376 421 -hsync -vsync interlaced)",
"arcade_15.75_928x348 vbeline(0 19.152 928 1000 1088 1216 348 372 376 421 -hsync -vsync interlaced)",
"arcade_15.75_480x360 vbeline(0 9.954 480 520 568 632 360 385 390 436 -hsync -vsync interlaced)",
"arcade_15.75_960x360 vbeline(0 19.908 960 1040 1128 1264 360 385 390 436 -hsync -vsync interlaced)",
"arcade_15.75_496x372 vbeline(0 10.206 496 528 576 648 372 398 402 450 -hsync -vsync interlaced)",
"arcade_15.75_992x372 vbeline(0 20.538 992 1072 1168 1304 372 398 402 450 -hsync -vsync interlaced)",
"arcade_15.75_512x384 vbeline(0 10.584 512 552 600 672 384 411 416 465 -hsync -vsync interlaced)",
"arcade_15.75_1024x384 vbeline(0 21.168 1024 1104 1200 1344 384 411 416 465 -hsync -vsync interlaced)",
"arcade_15.75_576x432 vbeline(0 11.97 576 624 680 760 432 462 468 523 -hsync -vsync interlaced)",
"arcade_15.75_1152x432 vbeline(0 23.814 1152 1240 1352 1512 432 462 468 523 -hsync -vsync interlaced)",
"arcade_15.75_640x480 vbeline(0 13.23 640 688 752 840 480 514 519 581 -hsync -vsync interlaced)",
"arcade_15.75_672x504 vbeline(0 13.86 672 720 784 880 504 539 545 610 -hsync -vsync interlaced)",

/***************************************************************************/
/* Arcade 24kHz */

/* computed */
"arcade_25_256x192 vbeline(0 8.4 256 272 304 336 192 195 197 208 -hsync -vsync doublescan)",
"arcade_25_512x192 vbeline(0 16.8 512 552 600 672 192 195 197 208 -hsync -vsync doublescan)",
"arcade_25_272x204 vbeline(0 9 272 296 320 360 204 207 209 221 -hsync -vsync doublescan)",
"arcade_25_544x204 vbeline(0 17.8 544 584 632 712 204 207 209 221 -hsync -vsync doublescan)",
"arcade_25_288x216 vbeline(0 9.4 288 312 336 376 216 219 222 234 -hsync -vsync doublescan)",
"arcade_25_576x216 vbeline(0 19 576 624 680 760 216 219 222 234 -hsync -vsync doublescan)",
"arcade_25_304x228 vbeline(0 10 304 328 360 400 228 232 234 247 -hsync -vsync doublescan)",
"arcade_25_608x228 vbeline(0 20 608 656 712 800 228 232 234 247 -hsync -vsync doublescan)",
"arcade_25_320x240 vbeline(0 10.6 320 344 376 424 240 244 246 260 -hsync -vsync)",
"arcade_25_640x240 vbeline(0 21 640 688 752 840 240 244 246 260 -hsync -vsync)",
"arcade_25_336x252 vbeline(0 11 336 360 392 440 252 256 259 273 -hsync -vsync)",
"arcade_25_672x252 vbeline(0 22 672 720 784 880 252 256 259 273 -hsync -vsync)",
"arcade_25_352x264 vbeline(0 11.6 352 384 416 464 264 268 271 286 -hsync -vsync)",
"arcade_25_704x264 vbeline(0 23.2 704 760 832 928 264 268 271 286 -hsync -vsync)",
"arcade_25_368x276 vbeline(0 12 368 392 432 480 276 280 283 299 -hsync -vsync)",
"arcade_25_384x288 vbeline(0 12.6 384 416 448 504 288 292 295 312 -hsync -vsync)",
"arcade_25_400x300 vbeline(0 13.2 400 432 472 528 300 305 308 325 -hsync -vsync)",
"arcade_25_416x312 vbeline(0 13.6 416 448 488 544 312 317 320 338 -hsync -vsync)",
"arcade_25_432x324 vbeline(0 14.2 432 464 504 568 324 329 332 351 -hsync -vsync)",
"arcade_25_448x336 vbeline(0 14.8 448 488 528 592 336 341 345 364 -hsync -vsync)",
"arcade_25_464x348 vbeline(0 15.2 464 496 544 608 348 353 357 377 -hsync -vsync)",
"arcade_25_480x360 vbeline(0 15.8 480 520 568 632 360 366 369 390 -hsync -vsync)",
"arcade_25_496x372 vbeline(0 16.2 496 528 576 648 372 378 382 403 -hsync -vsync)",
"arcade_25_512x384 vbeline(0 16.8 512 552 600 672 384 390 394 416 -hsync -vsync)",
"arcade_25_576x432 vbeline(0 19 576 624 680 760 432 439 443 468 -hsync -vsync)",
"arcade_25_640x480 vbeline(0 21 640 688 752 840 480 487 492 520 -hsync -vsync)",
"arcade_25_672x504 vbeline(0 22 672 720 784 880 504 512 517 546 -hsync -vsync)",

0
};

static int vbeline_search_target_mode(unsigned bits, unsigned req_x, unsigned req_y) {
	video_mode best;
	int best_set = 0;

	vbe_mode_iterator vmi;
	vbe_mode_iterator_begin(&vmi);
	while (!vbe_mode_iterator_end(&vmi)) {
		video_mode v;
		int found = 0;
		memset(&v,0,sizeof(v));

		v.driver.vbe.mode = vbe_mode_iterator_get(&vmi) | vbeLinearBuffer;
		found = vbe_mode_import(&v)==0;
		if (!found) {
			v.driver.vbe.mode = vbe_mode_iterator_get(&vmi);
			found = vbe_mode_import(&v)==0;
		}

		if (found) {
			if (v.size_x >= req_x && v.size_y >= req_y && v.bits == bits) {
				if (!best_set ||
					(v.size_x <= best.size_x && v.size_y <= best.size_y)) {
					best_set = 1;
					best = v;
				}
			}
		}

		vbe_mode_iterator_next(&vmi);
	}

	if (best_set)
		return best.driver.vbe.mode;
	else
		return -1;
}

video_error vbeline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	int mode_number;
	if (mode->driver.vbeline.mode != 0) {
		if (video_mode_import(mode)==0) {
			video_mode_container_insert(vmc,mode);
		}
	} else {
		video_mode mode8 = *mode;
		video_mode mode16 = *mode;

		strcat(mode8.name,"x8");
		mode_number = vbeline_search_target_mode(8,mode8.driver.vbeline.crtc.hde,mode8.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode8.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode8)==0) {
				video_mode_container_insert(vmc,&mode8);
			}
		}

		strcat(mode16.name,"x16");
		mode_number = vbeline_search_target_mode(16,mode16.driver.vbeline.crtc.hde,mode16.driver.vbeline.crtc.vde);
		if (mode_number >= 0) {
			mode16.driver.vbeline.mode = mode_number;
			if (video_mode_import(&mode16)==0) {
				video_mode_container_insert(vmc,&mode16);
			}
		}
	}

	return 0;
}

void video_mode_container_insert_vbeline(video_mode_container* vmc) {
	if (vbeline_active()) {
		const char** p = VBELINE_DEFAULT;
		while (*p) {
			video_mode mode;
			char* buffer = strdup(*p);
			memset(&mode,0,sizeof(mode));
			if (video_mode_parse_line(&mode,buffer)==0) {
				vbeline_mode_import_and_insert(vmc,&mode);
			}
			free(buffer);
			++p;
		}
	}
}

void video_mode_vbeline_print(char* s, const video_mode* mode) {
	s += sprintf(s,"%s vbeline(",mode->name);
	s += video_print_modeline_mc(s,mode->driver.vbeline.mode,&mode->driver.vbeline.crtc);
	s += sprintf(s,")");
}

void vbeline_wait_vsync(void) {
	if (vbe_mode_active())
		vbe_wait_vsync();
	else
		vga_wait_vsync();
}

int vbeline_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	if (vbe_mode_active())
		return vbe_palette8_set(palette,start,count,waitvsync);
	else {
		vga_palette8_set(palette,start,count,waitvsync);
		return 0;
	}
}
