/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "vga.h"

#include <stdio.h>

/* Import information for one video mode
	vm->driver.vga video mode
	Return 0 if successful
*/
int vga_mode_import(video_mode* mode) {
	if (mode->driver.vga.mode == 0x13) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_GRAPHICS |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 320;
		mode->size_y = 200;
		mode->virtual_x = 320;
		mode->virtual_y = 200;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 8;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x6A) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_INDEX_PACKED |
			VIDEO_FLAGS_TYPE_GRAPHICS |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 400;
		mode->size_y = 300;
		mode->virtual_x = 400;
		mode->virtual_y = 400;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 8;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x3) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_TEXT |
			VIDEO_FLAGS_TYPE_TEXT |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 80*9;
		mode->size_y = 25*16;
		mode->virtual_x = 80*9;
		mode->virtual_y = 25*16;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 0;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else if (mode->driver.vga.mode == 0x1) {
		mode->flags = VIDEO_FLAGS_DRIVER_VGA |
			VIDEO_FLAGS_LINEAR |
			VIDEO_FLAGS_INDEX_TEXT |
			VIDEO_FLAGS_TYPE_TEXT |
			(mode->flags & VIDEO_FLAGS_USER_MASK);
		mode->size_x = 40*9;
		mode->size_y = 25*16;
		mode->virtual_x = 80*9;
		mode->virtual_y = 25*16;
		mode->vert_freq = 0;
		mode->horz_freq = 0;
		mode->bits = 0;
		mode->aspect_x = 4 * mode->size_y;
		mode->aspect_y = 3 * mode->size_x;
	} else {
		video_error_description_set("Unsupported VGA mode %d",mode->driver.vga.mode);
		return -1;
	}

	return 0;
}

video_error vga_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

/* Add all vga mode */
void video_mode_container_insert_vga(video_mode_container* vmc) {
	video_mode v;

	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x13;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
		video_mode_container_insert(vmc,&v);
	}

	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x6A;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_%dx%dx%d",(unsigned)v.size_x,(unsigned)v.size_y,(unsigned)v.bits);
		video_mode_container_insert(vmc,&v);
	}

	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x3;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_text");
		video_mode_container_insert(vmc,&v);
	}

	memset(&v,0,sizeof(v));
	v.driver.vga.mode = 0x1;
	if (vga_mode_import(&v)==0) {
		sprintf(v.name,"bios_vga_text40");
		video_mode_container_insert(vmc,&v);
	}
}

void video_mode_vga_print(char* s, const video_mode* vm) {
	sprintf(s,"%s vga(%x)",vm->name,vm->driver.vga.mode);
}

/*
int vga_palette8_set(video_color* palette, unsigned start, unsigned count, int waitvsync) {
	video_color vga_pal[256];
	unsigned i;

	for(i=0;i<count;++i) {
		vga_pal[i].red = palette[i].red >> 2;
		vga_pal[i].green = palette[i].green >> 2;
		vga_pal[i].blue = palette[i].blue >> 2;
		vga_pal[i].alpha = 0;
	}
	vga_palette6_set(vga_pal,start,count,waitvsync);
}
*/
