/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "vgaline.h"
#include "vgatweak.h"

#include <stdlib.h>
#include <stdio.h>

static int vgaline_mode_13_realize(vgatweak_video_mode* regs, const video_crtc* crtc) {
	unsigned x_div = 4; /* 4 pixel for clock in 8 bit mode */
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* Get dot clock */
	unsigned dotclock = crtc->pixelclock * 8 / x_div;
	if (vga_dotclock_get(&dotclock) != 0)
		return -1;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(8, tweak_register_set, tweak_register_get, regs);

#if 0
	/* set char size to 1 */
	tweak_char_size_y_set(1, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);
#else
	/* set char size emulating doublescan */
	tweak_char_size_y_set(crtc->doublescan != 0 ? 2 : 1, tweak_register_set, tweak_register_get, regs);

	/* reset double scan */
	tweak_doublescan_set(0, tweak_register_set, tweak_register_get, regs);
#endif

	/* set dot clock */
	switch (dotclock) {
		case TWEAK_DOTCLOCK_HIGH:
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_HIGH/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		default:
			return -1;
	}

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	/* set chained/unchained mode */
	if (crtc->hde * crtc->vde > 0x10000) {
		tweak_chained_set(0,tweak_register_set, tweak_register_get, regs);
	} else {
		tweak_chained_set(1,tweak_register_set, tweak_register_get, regs);
	}

	return 0;
}

static int vgaline_mode_6A_realize(vgatweak_video_mode* regs, const video_crtc* crtc) {
	unsigned x_div = 4; /* 4 pixel for clock in 8 bit mode */
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(8, tweak_register_set, tweak_register_get, regs);

#if 0
	/* set char size to 1 */
	tweak_char_size_y_set(1, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);
#else
	/* set char size emulating doublescan */
	tweak_char_size_y_set(crtc->doublescan != 0 ? 2 : 1, tweak_register_set, tweak_register_get, regs);

	/* reset double scan */
	tweak_doublescan_set(0, tweak_register_set, tweak_register_get, regs);
#endif

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	return 0;
}

static int vgaline_mode_text_realize(vgatweak_video_mode* regs, const video_crtc* crtc, unsigned font_x, unsigned font_y) {
	unsigned x_div = font_x;
	unsigned y_mul = crtc->doublescan!=0 ? 2 : 1;
	unsigned de,bs,rs,re,be,tt;

	/* Get dot clock (for text mode is equal at dot clock) */
	unsigned dotclock = crtc->pixelclock;
	if (vga_dotclock_get(&dotclock) != 0)
		return -1;

	/* output horizontal */
	video_crtc_realize_h(crtc,&de,&bs,&rs,&re,&be,&tt,x_div);
	tweak_htt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_hde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_hrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_hre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_hbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_hbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* output vertical */
	video_crtc_realize_v(crtc,&de,&bs,&rs,&re,&be,&tt,y_mul);
	tweak_vtt_set(tt, tweak_register_set, tweak_register_get, regs);
	tweak_vde_set(de, tweak_register_set, tweak_register_get, regs);
	tweak_vrs_set(rs, tweak_register_set, tweak_register_get, regs);
	tweak_vre_set(re, tweak_register_set, tweak_register_get, regs);
	tweak_vbs_set(bs, tweak_register_set, tweak_register_get, regs);
	tweak_vbe_set(be, tweak_register_set, tweak_register_get, regs);

	/* set char size */
	tweak_char_size_x_set(font_x, tweak_register_set, tweak_register_get, regs);
	tweak_char_size_y_set(font_y, tweak_register_set, tweak_register_get, regs);

	/* set double scan as requested */
	tweak_doublescan_set(crtc->doublescan != 0, tweak_register_set, tweak_register_get, regs);

	/* set dot clock */
	switch (dotclock) {
		case TWEAK_DOTCLOCK_HIGH :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_HIGH/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(1, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW :
			tweak_dotclock_middle_set(0, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		case TWEAK_DOTCLOCK_LOW/2 :
			tweak_dotclock_middle_set(1, tweak_register_set, tweak_register_get, regs);
			tweak_masterclock_input_set(0, tweak_register_set, tweak_register_get, regs);
		break;
		default:
			return -1;
	}

	/* set polarity */
	tweak_hsync_set(crtc->hsync, tweak_register_set, tweak_register_get, regs);
	tweak_vsync_set(crtc->vsync, tweak_register_set, tweak_register_get, regs);

	/* set scanline length */
	tweak_offset_set((crtc->hde + 2 * x_div - 1) / (2 * x_div), tweak_register_set, tweak_register_get, regs);

	return 0;
}

int vgaline_mode_realize(vgatweak_video_mode* regs, const vgaline_video_mode* mode) {
	unsigned font;
	switch (mode->mode) {
		case 0x13:
			*regs = tweak_13_default;
			if (vgaline_mode_13_realize(regs,&mode->crtc)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x13");
				return -1;
			}
			break;
		case 0x6A:
			*regs = tweak_6A_default;
			if (vgaline_mode_6A_realize(regs,&mode->crtc)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x6A");
				return -1;
			}
			break;
		case 0x1 :
			*regs = tweak_1_default;
			if (mode->crtc.hde >= 360)
				font = 9;
			else
				font = 8;
			if (vgaline_mode_text_realize(regs,&mode->crtc,font,8)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x1");
				return -1;
			}
			break;
		case 0x3 :
			*regs = tweak_3_default;
			if (mode->crtc.hde >= 720)
				font = 9;
			else
				font = 8;
			if (vgaline_mode_text_realize(regs,&mode->crtc,font,8)!=0) {
				video_error_description_set("Error in vgaline mode realize 0x3");
				return -1;
			}
			break;
		default:
			video_error_description_set("Unsupported mode %02x in vgaline mode realize", (unsigned)mode->mode);
			return -1;
	}
	return 0;
}

int vgaline_mode_set(vgaline_video_mode* mode) {
	vgatweak_video_mode regs;

	video_log_modeline_m("vgaline: mode_set", mode->mode, mode->crtc.pixelclock, mode->crtc.hde, mode->crtc.hrs, mode->crtc.hre, mode->crtc.ht, mode->crtc.vde, mode->crtc.vrs, mode->crtc.vre, mode->crtc.vt, mode->crtc.hsync, mode->crtc.vsync, mode->crtc.doublescan, mode->crtc.interlaced );

 	if (vgaline_mode_realize(&regs,mode)!=0) {
		return -1;
	}

	if (vga_mode_set(mode->mode,regs.regs,regs.regs_num)!=0) {
		return -1;
	}

	return 0;
}

void vgaline_mode_done(void) {
	vga_mode_done();
}

int vgaline_mode_grab(vgaline_video_mode* mode) {
	struct tweak_crtc info;
	video_mode dummy;
	dummy.flags = VIDEO_FLAGS_DRIVER_VGALINE;

	if (!vga_active())
		return -1;

	/* get the video mode */
	if (vga_mode_get(&dummy.driver.vgaline.mode) != 0)
		return -1;

	/* import information */
	tweak_crtc_get(&info,tweak_reg_read,0);

	/* generate crtc values */
	if (video_crtc_import(&dummy.driver.vgaline.crtc,&info,&dummy,info.vert_clock)!=0)
		return -1;

	*mode = dummy.driver.vgaline;

	return 0;
}

/* Import information for one video mode
	Return 0 if successful
*/
int vgaline_mode_import(video_mode* mode) {
	vgatweak_video_mode regs;

	/* adjust the pixel clock to an acceptable value */
	if (vga_pixelclock_get(&mode->driver.vgaline.crtc.pixelclock,mode->driver.vgaline.mode) != 0)
		return -1;

	if (vgaline_mode_realize(&regs,&mode->driver.vgaline)!=0) {
		return -1;
	}

	if (vgatweak_mode_import_ext(mode,&regs)!=0) {
		return -1;
	}

	/* set values that depend on crtc values */
	mode->size_x = mode->driver.vgaline.crtc.hde;
	mode->size_y = mode->driver.vgaline.crtc.vde;
	mode->horz_freq = (double)mode->driver.vgaline.crtc.pixelclock / mode->driver.vgaline.crtc.ht;
	mode->vert_freq =  mode->horz_freq / mode->driver.vgaline.crtc.vt;
	if (mode->driver.vgaline.crtc.doublescan)
		mode->vert_freq /= 2;
	if (mode->driver.vgaline.crtc.interlaced)
		mode->vert_freq *= 2;

	mode->flags = (mode->flags & ~VIDEO_FLAGS_DRIVER_MASK) | VIDEO_FLAGS_DRIVER_VGALINE;

	return 0;
}

static const char* VGALINE_DEFAULT[] = {

/***************************************************************************/
/* PC MONITOR AT 31.5 kHz */

"pc_31.5_vga_304x228x8 vgaline(13 12.59 304 324 356 400 228 236 238 245 doublescan -hsync -vsync)",
"pc_31.5_vga_320x240x8 vgaline(13 12.59 320 336 356 400 240 249 254 262 doublescan -hsync -vsync)",
"pc_31.5_vga_336x252x8 vgaline(13 14.16 336 368 416 448 252 266 270 288 doublescan -hsync -vsync)",
"pc_31.5_vga_384x288x8 vgaline(13 14.16 384 392 408 448 288 290 292 298 doublescan -hsync -vsync)",

/* TEXT resolutions */
"pc_31.5_vga_text vgaline(3 28.32 720 765 873 900 400 412 414 449 -hsync +vsync)",

/***************************************************************************/
/* ARCADE 15 kHz */

/* TEXT resolutions, from ArcadeOS */
"arcade_15.75_vga_text vgaline(3 14.16 640 712 792 880 200 225 227 267 -hsync -vsync)",

/***************************************************************************/
/* PAL TV */

/* MAX resolutions, computed */
"pal_vga_736x288 vgaline(13 14.161 736 760 824 904 288 291 294 312 -hsync -vsync)",
"pal_vga_368x288 vgaline(13 7.0805 368 376 416 456 288 291 294 312 -hsync -vsync)",
"pal_vga_656x288 vgaline(13 12.5875 656 672 736 808 288 291 294 312 -hsync -vsync)",
"pal_vga_328x288 vgaline(13 6.29375 328 336 368 400 288 291 294 312 -hsync -vsync)",

/* SMALL resolutions, manual */
"pal_small_vga_304x276x8 vgaline(13 6.2938 304 320 372 400 276 284 285 312 -hsync -vsync)",
"pal_small_vga_336x276x8 vgaline(13 7.0805 336 356 404 452 276 284 285 312 -hsync -vsync)",

/* STANDARD resolutions, computed */
"pal_std_vga_320x200 vgaline(13 6.29375 320 336 360 400 200 247 250 312 -hsync -vsync)",
"pal_std_vga_320x240 vgaline(13 6.29375 320 336 360 400 240 267 270 312 -hsync -vsync)",

/* from VGA2PAL */
/*"pal_vga_text_vga2pal vgaline(3 14.16 643 700 828 896 256 276 280 311 -hsync -vsync)",*/

/* TEXT resolutions, computed */
"pal_vga_text vgaline(3 12.5875 640 672 728 808 200 247 250 312 -hsync -vsync)",

/***************************************************************************/
/* NTSC TV */

/* MAX resolutions, computed */
"ntsc_vga_744x242 vgaline(13 14.161 744 768 832 904 242 245 248 262 -hsync -vsync)",
"ntsc_vga_376x242 vgaline(13 7.0805 376 384 416 448 242 245 248 262 -hsync -vsync)",
"ntsc_vga_664x242 vgaline(13 12.5875 664 680 744 800 242 245 248 262 -hsync -vsync)",
"ntsc_vga_328x242 vgaline(13 6.29375 328 344 368 400 242 245 248 262 -hsync -vsync)",

/* SMALL resolutions, manual */
"ntsc_small_vga_304x228 vgaline(13 6.2938 304 320 372 400 228 235 238 262 -hsync -vsync)",
"ntsc_small_vga_336x228 vgaline(13 7.0805 336 356 404 452 228 235 238 262 -hsync -vsync)",

/* STANDARD resolutions, computed */
"ntsc_std_vga_320x200 vgaline(13 6.29375 320 336 368 400 200 224 227 262 -hsync -vsync)",
"ntsc_std_vga_320x240 vgaline(13 6.29375 320 336 368 400 240 244 247 262 -hsync -vsync)",

/* from VGA2NTSC */
/*"ntsc_vga_text_vga2ntsc vgaline(3 14.16 640 736 800 904 200 220 229 263 -hsync -vsync)",*/

/* TEXT resolutions, computed */
"ntsc_vga_text vgaline(3 12.5875 640 672 728 800 200 224 227 262 -hsync -vsync)",

/***************************************************************************/
/* ARCADE 25 kHz */

/* MAX resolutions, computed */
"arcade_25_vga_432x384 vgaline(13 14.161 432 464 504 568 384 390 394 416 -hsync -vsync)",
"arcade_25_vga_216x384 vgaline(13 7.0805 216 232 256 280 384 390 394 416 -hsync -vsync)",
"arcade_25_vga_384x384 vgaline(13 12.5875 384 416 448 504 384 390 394 416 -hsync -vsync)",
"arcade_25_vga_192x384 vgaline(13 6.29375 192 208 224 248 384 390 394 416 -hsync -vsync)",

/* STANDARD resolutions, computed */
"arcade_25_vga_320x200 vgaline(13 12.5875 320 384 416 504 200 298 302 416 -hsync -vsync)",
"arcade_25_vga_320x240 vgaline(13 12.5875 320 384 416 504 240 318 322 416 -hsync -vsync)",

/* TEXT resolutions, computed */
"arcade_25_vga_text vgaline(1 12.5875 320 384 416 504 200 298 302 416 -hsync -vsync)",

0
};

video_error vgaline_mode_import_and_insert(video_mode_container* vmc, video_mode* mode) {
	video_mode mode8 = *mode;
	if (video_mode_import(&mode8)==0) {
		video_mode_container_insert(vmc,&mode8);
	}
	return 0;
}

void video_mode_container_insert_vgaline(video_mode_container* vmc) {
	const char** p = VGALINE_DEFAULT;
	while (*p) {
		video_mode mode;
		char* buffer = strdup(*p);
		memset(&mode,0,sizeof(mode));
		if (video_mode_parse_line(&mode,buffer)==0) {
			vgaline_mode_import_and_insert(vmc,&mode);
		}
		free(buffer);
		++p;
	}
}

void video_mode_container_insert_vgaline_only_text(video_mode_container* vmc) {
	const char** p = VGALINE_DEFAULT;
	while (*p) {
		video_mode mode;
		char* buffer = strdup(*p);
		memset(&mode,0,sizeof(mode));
		if (video_mode_parse_line(&mode,buffer)==0) {
			if (mode.driver.vgaline.mode == 0x3)
				vgaline_mode_import_and_insert(vmc,&mode);
		}
		free(buffer);
		++p;
	}
}

void video_mode_container_insert_vgaline_only_standard(video_mode_container* vmc) {
	const char** p = VGALINE_DEFAULT;
	while (*p) {
		video_mode mode;
		char* buffer = strdup(*p);
		memset(&mode,0,sizeof(mode));
		if (video_mode_parse_line(&mode,buffer)==0) {
			if ((mode.driver.vgaline.crtc.hde == 320 && mode.driver.vgaline.crtc.vde == 200)
				|| (mode.driver.vgaline.crtc.hde == 320 && mode.driver.vgaline.crtc.vde == 240)
			) {
				vgaline_mode_import_and_insert(vmc,&mode);
			}
		}
		free(buffer);
		++p;
	}
}

void video_mode_vgaline_print(char* s, const video_mode* mode) {
	s += sprintf(s,"%s vgaline(",mode->name);
	s += video_print_modeline_mc(s,mode->driver.vgaline.mode,&mode->driver.vgaline.crtc);
	s += sprintf(s,")");
}

